# Commands
#
# Slides args appear on !SLIDE line. E.g.
#
#   !SLIDE_HERE myclass
#
# creates a slide in current pos with class myclass (in the HTML)
#
# Use scale to set zoom, this will be applied after next slide created.
# First arg is movement scale, second is size of slides. Scale relative
# does it wrt to current zoom
#
# !SCALE .5 .3
# !SCALE_RELATIVE .5 .3
#
# Set next slides to .5 size and zoom in
#
# !SLIDE_HERE args
#
# Create a new slide in the current position (first slide).
# SLIDE_SCALE_LEFT allows the distance left to be scaled (e.g. half
# normal distance)
#
# !SLIDE_RIGHT args
# !SLIDE_LEFT args
# !SLIDE_DOWN args
# !SLIDE_SCALE_LEFT .5 args
#
# Move one slide unit left and create a new slide
#
# !SLIDE_MOVE x y
#
# Move x units horizontal and y units vertical
#
# !SLIDE_RELATIVE rx ry
#
# Useful for making slides within slides, e.g.
#
#   !SCALE .25
#   !SLIDE_RELATIVE .25 -.25
#
# will make a 1/4 size sub-slide a quarter of the way from the top/left
#
# !VSPACE .2
#
# Add vspace of .2 of the slide height
#
# !PUSH
#
# Save current position
#
# !POP
#
# Return to last saved position
#
# To make rows of slides with titles, yoffset is optional and used when
# the sliderows have material before them
#
# !START_SLIDEROWS 2 <yoffset>
# !ROWTITLE title1
# !ROWTITLE title2
# !START_SLIDEROW 3
# !SLIDE_HERE
# slide 1
# !SLIDE_LEFT
# slide 2
# !SLIDE_LEFT
# slide 3
# !END_SLIDEROW
# !START_SLIDEROW 3
# ...
# !END_SLIDEROW
# !END_SLIDEROWS
#
# To make a slide step
#
# appears as normal
# {% slide_step %}
# appears later!
# {% end %}
# appears as normal
#
# To add highlighted code
#
# {% embed_code ruby %}
#   def blah()
#
#   end
# {% end %}
#
# Embedded code will be surrounded by <div class="code_block"> for your styling.
#
# Images can be added with ![](path/to/image) but HTML can be used for
# more control
#
# Repl.it can be embedded with plus optional height (in HTML format)
#
# !REPL 500px @matthewhague/EmptyPythonRepl
#
# Similar for general iframes
#
# !IFRAME 500px <url>
#
# Import a whole file verbatim
#
# !IMPORT <filename>
#
# Bookmark and return to a location
#
# !BOOKMARK <label>
# !GOTO <label>

require 'pygments'

module SlideHelper
    @@HUNIT = 1000.0
    @@VUNIT = 800.0
    @@OVERVIEW_VUNIT = 9 * @@HUNIT / 16

    @@max_x = 0
    @@min_x = 0
    @@max_y = 0
    @@min_y = 0

    @@next_iframe_id = 0

    class SlidePosition
        attr_accessor :x, :y,
                      :mov_scale, :slide_scale,
                      :next_mov_scale, :next_slide_scale
        def initialize()
            @x = 0
            @y = 0
            @mov_scale = 1
            @slide_scale = 1
            @next_mov_scale = 1
            @next_slide_scale = 1
        end

        def to_s
          "x = #{@x}, y = #{@y}"
        end
    end

    @@pos = SlidePosition.new
    @@stack = []
    @@bookmarks = { }

    class SlideRowsData
        attr_accessor :num, :cur_row, :yoffset
        def initialize(num_rows = 0, yoffset = 0)
            @num = num_rows
            @cur_row = 0
            @yoffset = yoffset
        end
    end

    @@cur_sliderows = SlideRowsData.new
    @@sliderows_stack = []

    def slide_step(opts={}, &blk)
        text = capture_erb(&blk)

        before = "<div class='step inner'>\n"
        after  = "</div>\n"

        html  = ""
        html << guard_block(before)
        html << markdown_to_html(text)
        html << guard_block(after)

        concat_erb(html, blk.binding)
    end

    def embed_code(opts="java", &blk)
        code = capture_erb(&blk)

        language = opts

        before = "<div class='code_block'>\n"
        after  = "</div>\n"
        code = Pygments.highlight(code, {
          :lexer => language,
          :options => { :noclasses => true }
        })

        html  = ""
        html << guard_block(before)
        html << guard_block(code)
        html << guard_block(after)

        concat_erb(html, blk.binding)
    end


    def impress_demote_headers(input)
        input.gsub(/^#/, "###")
    end

    def impress_layout_slides(input)
        output = ""
        input.each_line { |line|
            cmd, args = impress_break_line(line)
            case line.split.first
            when '!SCALE'
                output += impress_scale(args) + "\n"
            when '!SCALE_RELATIVE'
                output += impress_scale_relative(args) + "\n"
            when '!SLIDE_HERE'
                output += impress_make_slide(args) + "\n"
            when '!SLIDE_RIGHT'
                output += impress_slide_right(args) + "\n"
            when '!SLIDE_LEFT'
                output += impress_slide_left(args) + "\n"
            when '!SLIDE_DOWN'
                output += impress_slide_down(args) + "\n"
            when '!SLIDE_UP'
                output += impress_slide_up(args) + "\n"
            when '!SLIDE_MOVE'
                output += impress_slide_move(args) + "\n"
            when '!SLIDE_SCALE_LEFT'
                output += impress_slide_scale_left(args) + "\n"
            when '!SLIDE_RELATIVE'
                output += impress_slide_relative(args) + "\n"
            when '!VSPACE'
                output += impress_vspace(args) + "\n"
            when '!PUSH'
                output += impress_push(args) + "\n"
            when '!POP'
                output += impress_pop(args) + "\n"
            when '!START_SLIDEROWS'
                output += impress_start_sliderows(args) + "\n"
            when '!END_SLIDEROWS'
                output += impress_end_sliderows(args) + "\n"
            when '!ROWTITLE'
                output += impress_row_title(args) + "\n"
            when '!START_SLIDEROW'
                output += impress_start_sliderow(args) + "\n"
            when '!END_SLIDEROW'
                output += impress_end_sliderow(args) + "\n"
            when '!OVERVIEW'
                output += impress_overview(args) + "\n"
            when '!REPL'
                output += impress_repl(args) + "\n"
            when '!IFRAME'
                output += impress_iframe(args) + "\n"
            when '!IMPORT'
                output += impress_import(args) + "\n"
            when '!BOOKMARK'
                output += impress_bookmark(args) + "\n"
            when '!GOTO'
                output += impress_goto(args) + "\n"
            else
                output += line
            end
        }

        output
    end

    def impress_push(args = "")
        @@stack.push(@@pos.clone)
        ''
    end

    def impress_pop(args = "")
        @@pos = @@stack.pop()
        "!SLIDE unclickable x=#{@@pos.x} y=#{@@pos.y} scale=#{@@pos.slide_scale} #{args}"
    end

    def impress_bookmark(args)
      puts "Bookmark #{args} #{@@pos}"
      @@bookmarks[args] = @@pos.clone
      ''
    end

    def impress_goto(args)
        @@pos = @@bookmarks[args].clone
        "!SLIDE unclickable x=#{@@pos.x} y=#{@@pos.y} scale=#{@@pos.slide_scale} #{args}"
    end

    def impress_scale(args)
        argv = args.split
        @@pos.next_mov_scale = argv[0].to_f
        @@pos.next_slide_scale = argv[1].to_f
        ""
    end

    def impress_scale_relative(args)
        argv = args.split
        @@pos.next_mov_scale *= argv[0].to_f
        @@pos.next_slide_scale *= argv[1].to_f
        ""
    end

    def impress_slide_right(args)
        @@pos.x -= impress_horiz_mov_unit()
        impress_make_slide(args)
    end

    def impress_slide_left(args)
        @@pos.x += impress_horiz_mov_unit()
        impress_make_slide(args)
    end

    def impress_slide_down(args)
        @@pos.y += impress_vert_mov_unit()
        impress_make_slide(args)
    end

    def impress_slide_up(args)
        @@pos.y -= impress_vert_mov_unit()
        impress_make_slide(args)
    end

    def impress_slide_move(args)
        argv = args.split
        rx, ry = argv[0..1].map {|v| v.to_f}
        rem_args = argv[2..-1].join(' ')
        @@pos.x += rx * impress_horiz_mov_unit()
        @@pos.y += ry * impress_vert_mov_unit()
        impress_make_slide(rem_args)
    end

    def impress_slide_scale_left(args)
        argv = args.split
        scale = argv[0].to_f
        rem_args = argv[1..-1].join(' ')
        @@pos.x += scale * impress_horiz_mov_unit()
        impress_make_slide(rem_args)
    end

    def impress_slide_relative(args)
        argv = args.split
        rx, ry = argv[0..1].map {|v| v.to_f}
        # move relative to current slide (not movement unit)
        x_delta = (rx - 0.5) * impress_horiz_slide_unit
        y_delta = (ry - 0.5) * impress_vert_slide_unit
        @@pos.x += x_delta
        @@pos.y += y_delta
        impress_make_slide(argv[2..-1].join(' '))
    end

    def impress_make_slide(args)
        @@pos.mov_scale = @@pos.next_mov_scale
        @@pos.slide_scale = @@pos.next_slide_scale

        @@max_x = [@@max_x, @@pos.x].max
        @@min_x = [@@min_x, @@pos.x].min
        @@max_y = [@@max_y, @@pos.y].max
        @@min_y = [@@min_y, @@pos.y].min

        "!SLIDE #{args} x=#{@@pos.x} y=#{@@pos.y} scale=#{@@pos.slide_scale}"
    end

    def impress_overview(args)
        mid_x = (@@max_x + @@min_x) / 2
        mid_y = (@@max_y + @@min_y) / 2
        scale_x = (@@max_x - @@min_x) / @@HUNIT
        scale_y = (@@max_y - @@min_y) / @@OVERVIEW_VUNIT
        scale = [scale_x, scale_y].max
        "!SLIDE overview #{args} x=#{mid_x} y=#{mid_y} scale=#{scale}"
    end

    def impress_vspace(args)
        pix = args.to_f * impress_vert_slide_unit
        "<p style='height: #{pix}px'/>"
    end

    def impress_start_sliderows(args)
        @@sliderows_stack.push(@@cur_sliderows.clone)

        argv = args.split
        num_rows = argv[0].to_i
        yoffset = 0
        if argv.length > 1
            yoffset = argv[1].to_f
        end

        @@cur_sliderows = SlideRowsData.new(num_rows, yoffset)
        ""
    end

    def impress_end_sliderows(args)
        @@cur_sliderows = @@sliderows_stack.pop
        ""
    end

    def impress_row_title(args)
        offset_factor = (1 - @@cur_sliderows.yoffset)
        # no row takes up more than 1/3 of slide space
        effective_rows = [3,@@cur_sliderows.num].max
        vspace = 0.9 / effective_rows * @@VUNIT * offset_factor
        "<h3 class='rowtitle' style='height: #{vspace}px'>#{args}&nbsp;</h3>"
    end

    def impress_start_sliderow(args)
        argv = args.split
        num_slides = argv[0].to_i
        slide_scale_factor = 1
        if (argv.length > 1)
            slide_scale_factor = argv[1].to_f
        end

        hscale = 1.0 / (num_slides + 1)
        vscale = 0.9 / @@cur_sliderows.num

        mov_scale = [hscale, vscale].min * @@pos.slide_scale

        slide_vscale = 0.5 * vscale
        slide_scale = slide_scale_factor * [hscale, slide_vscale].min * @@pos.slide_scale

        # do this before messing with pos x and y
        output = impress_push + "\n" + impress_scale("#{mov_scale} #{slide_scale}")

        x_delta = (hscale - 0.5) * impress_horiz_slide_unit
        fudged_yoffset = @@cur_sliderows.yoffset
        if fudged_yoffset > 0
            fudged_yoffset -= 0.4
        end
        y_baseline = (fudged_yoffset - 0.5) * impress_vert_slide_unit
        y_offset = (@@cur_sliderows.cur_row + 0.7) * vscale * impress_vert_slide_unit
        y_delta = y_baseline + y_offset
        @@pos.x += x_delta
        @@pos.y += y_delta

        @@cur_sliderows.cur_row += 1

        output
    end

    def impress_end_sliderow(args)
        impress_pop(args)
    end

    def impress_horiz_mov_unit()
        @@HUNIT * @@pos.mov_scale
    end

    def impress_vert_mov_unit()
        @@VUNIT * @@pos.mov_scale
    end

    def impress_horiz_slide_unit()
        @@HUNIT * @@pos.slide_scale
    end

    def impress_vert_slide_unit()
        @@VUNIT * @@pos.slide_scale
    end

    def impress_break_line(line)
        sline = line.split
        cmd = sline[0]
        args = ""
        if sline.length > 1
            args = sline[1..-1].join(" ")
        end
        [cmd, args]
    end

    def impress_repl(args)
        argv = args.split
        name = argv[-1]
        argv[-1] = "https://repl.it/#{name}?lite=true"
        impress_iframe(argv.join(" "))
    end

    def impress_iframe(args)
        argv = args.split
        url = ""
        height = ""
        if argv.length == 1
            height = '500px'
            url = argv[0]
        else
            height = argv[0]
            url = argv[1]
        end

        container_id = "iframecontainer#{@@next_iframe_id}"
        iframe_id = "iframe#{@@next_iframe_id}"
        @@next_iframe_id += 1

        # apologies to coding decency..
        onclick_js = "document.getElementById(\"#{iframe_id}\").style = \"position:absolute;left:50%;top:50%;transform:translate(-50%,-50%);width:95vw;height:95vh;border:1px dashed black\";"
        onclick_js += "document.getElementById(\"#{container_id}\").style = \"position:absolute;left:50%;top:50%;transform:translate(-50%,-50%);width:100vw;height:100vh;background:#efefefaa;cursor:pointer\""
        onoff_js = "document.getElementById(\"#{iframe_id}\").style = \"width:100%\";"
        onoff_js += "document.getElementById(\"#{container_id}\").style = \"\""

        ("<div onclick='#{onoff_js}' id='#{container_id}'>" +
         "<iframe id='#{iframe_id}' onload='blur()' frameborder='0' width='100%' height='#{height}' src='#{url}'></iframe>" +
         "</div>" +
         "<p style='width:100%;text-align:right;font-size:75%;cursor:pointer' onclick='#{onclick_js}'>↗</p>")
    end

    def impress_import(args)
      # not memory efficient
      impress_layout_slides(File.open(args, 'r').read)
    end
end

class Slideshow::Gen
    include SlideHelper
end
